from pathlib import Path

from yet_another_imod_wrapper.fiducials import generate_fiducial_alignment_directive


def test_alignment_directive_generation():
    """Check that fiducial based alignment directive is properly generated."""
    directive = generate_fiducial_alignment_directive(
        tilt_series_file=Path('test.mrc'),
        pixel_size=1.05,
        fiducial_size=10,
        rotation_angle=85.5
    )
    assert directive == {
        'setupset.copyarg.userawtlt': '1',
        'setupset.copyarg.stackext': '.mrc',
        'setupset.copyarg.rotation': str(85.5),
        'setupset.copyarg.pixel': str(1.05 / 10),
        'setupset.copyarg.gold': str(10),
        'setupset.copyarg.dual': '0',
        'runtime.Positioning.any.wholeTomogram': '1',
        'runtime.Fiducials.any.trackingMethod': '0',
        'runtime.Fiducials.any.seedingMethod': '3',
        'runtime.Excludeviews.any.deleteOldFiles': '0',
        'runtime.AlignedStack.any.binByFactor': '8',
        'comparam.prenewst.newstack.BinByFactor': str(8),
        'comparam.prenewst.newstack.AntialiasFilter': '-1',
        'comparam.autofidseed.autofidseed.TargetNumberOfBeads': '50',
        'comparam.track.beadtrack.SobelFilterCentering': '1',
        'comparam.track.beadtrack.ScalableSigmaForSobel': '0.12',
        'comparam.newst.newstack.TaperAtFill': '1,1',
        'comparam.newst.newstack.AntialiasFilter': '-1',
        'comparam.golderaser.ccderaser.ExpandCircleIterations': '3',
        'comparam.eraser.ccderaser.PeakCriterion': '8.0',
        'comparam.eraser.ccderaser.DiffCriterion': '6.0',
        'comparam.align.tiltalign.TiltOption': '0',
        'comparam.align.tiltalign.SurfacesToAnalyze': '1',
        'comparam.align.tiltalign.RotOption': '-1',
        'comparam.align.tiltalign.MagOption': '0',
        'comparam.align.tiltalign.BeamTiltOption': '2'
    }



import numpy as np

from yet_another_imod_wrapper.utils import (
    find_optimal_power_of_2_binning_factor,
    find_optimal_integer_binning_factor,
    prepare_imod_directory
)


def test_find_optimal_power_of_2_binning_factor():
    result = find_optimal_power_of_2_binning_factor(
        src_pixel_size=2, target_pixel_size=10
    )
    assert result == 4


def test_find_optimal_integer_binning_factor():
    result = find_optimal_integer_binning_factor(
        src_pixel_size=3, target_pixel_size=10,
    )
    assert result == 3


def test_prepare_imod_directory(tmp_path):
    imod_directory = tmp_path / 'imod'
    tilt_series_file = tmp_path / 'my_tilt_series.mrc'
    tilt_series_file.touch()
    tilt_angles = np.arange(-6, 9, 3)

    root_name = tilt_series_file.stem
    prepare_imod_directory(
        tilt_series_file=tilt_series_file,
        tilt_angles=tilt_angles,
        imod_directory=imod_directory
    )

    output_rawtlt_file = imod_directory / f'{root_name}.rawtlt'
    output_tilt_series = imod_directory / tilt_series_file.name
    assert output_rawtlt_file.exists()
    assert output_tilt_series.exists()

    output_tilt_angles = np.loadtxt(output_rawtlt_file)
    assert np.allclose(tilt_angles, output_tilt_angles)



from pathlib import Path

import numpy as np

from yet_another_imod_wrapper.fiducials import run_fiducial_based_alignment

TEST_DATA_DIR = Path(__file__).parent.parent / 'tilt_series'

run_fiducial_based_alignment(
    tilt_series_file=TEST_DATA_DIR / 'TS_01.mrc',
    tilt_angles=np.arange(-60, 63, 3),
    pixel_size=1.35,
    fiducial_size=10,
    nominal_rotation_angle=85,
    output_directory=Path('test_TS_01_fiducials')
)



from os import PathLike
from typing import Dict


def read_adoc(adoc_file: PathLike) -> Dict[str, str]:
    with open(adoc_file) as adoc:
        lines = adoc.readlines()
    lines = [
        line.strip().split('=')
        for line in lines
        if not line.startswith('#')
    ]
    lines = [line for line in lines if len(line) == 2]
    return {k.strip(): v.strip() for k, v in lines}


def write_adoc(data: dict, output_filename: PathLike):
    with open(output_filename, 'w') as adoc:
        for k, v in data.items():
            adoc.write(f"{k} = {v}\n")



from pathlib import Path

from yet_another_imod_wrapper.patch_tracking import generate_patch_tracking_alignment_directive


def test_alignment_directive_generation():
    """Check that patch tracking directive is properly generated."""
    directive = generate_patch_tracking_alignment_directive(
        tilt_series_file=Path('test.mrc'),
        pixel_size=1.05,
        rotation_angle=45.5,
        patch_size_xy=(50, 100),
        patch_overlap_percentage=33
    )
    assert directive == {
        'setupset.copyarg.userawtlt': '1', 'setupset.copyarg.stackext': '.mrc',
        'setupset.copyarg.rotation': str(45.5),
        'setupset.copyarg.pixel': str(1.05 / 10),
        'setupset.copyarg.gold': '10', 'setupset.copyarg.dual': '0',
        'runtime.Positioning.any.wholeTomogram': '1',
        'runtime.PatchTracking.any.adjustTiltAngles': '0',
        'runtime.Fiducials.any.trackingMethod': '1',
        'runtime.Fiducials.any.seedingMethod': '3',
        'runtime.Excludeviews.any.deleteOldFiles': '0',
        'runtime.AlignedStack.any.binByFactor': '8',
        'comparam.xcorr_pt.tiltxcorr.SizeOfPatchesXandY': '6,12',
        'comparam.xcorr_pt.tiltxcorr.OverlapOfPatchesXandY': '0.33,0.33',
        'comparam.xcorr_pt.tiltxcorr.ShiftLimitsXandY': '200,200',
        'comparam.xcorr_pt.tiltxcorr.LengthAndOverlap': '16,4',
        'comparam.xcorr_pt.tiltxcorr.IterateCorrelations': '20',
        'comparam.xcorr_pt.tiltxcorr.FilterSigma2': '0.03',
        'comparam.xcorr_pt.tiltxcorr.FilterRadius2': '0.125',
        'comparam.track.beadtrack.SobelFilterCentering': '1',
        'comparam.track.beadtrack.ScalableSigmaForSobel': '0.12',
        'comparam.tilt.tilt.THICKNESS': '3000',
        'comparam.prenewst.newstack.BinByFactor': str(8),
        'comparam.prenewst.newstack.AntialiasFilter': '-1',
        'comparam.newst.newstack.TaperAtFill': '1,1',
        'comparam.newst.newstack.AntialiasFilter': '-1',
        'comparam.golderaser.ccderaser.ExpandCircleIterations': '3',
        'comparam.eraser.ccderaser.PeakCriterion': '8.0',
        'comparam.eraser.ccderaser.DiffCriterion': '6.0',
        'comparam.align.tiltalign.TiltOption': '0',
        'comparam.align.tiltalign.SurfacesToAnalyze': '1',
        'comparam.align.tiltalign.RotOption': '-1',
        'comparam.align.tiltalign.MagOption': '0',
        'comparam.align.tiltalign.BeamTiltOption': '2'
    }



from os import PathLike
from pathlib import Path
from typing import Dict, Any, List

from .batchruntomo_config.io import read_adoc
from .constants import TARGET_PIXEL_SIZE_FOR_ALIGNMENT, BATCHRUNTOMO_CONFIG_FIDUCIALS
from .utils import (
    find_optimal_power_of_2_binning_factor,
    prepare_imod_directory,
    run_batchruntomo,
    imod_is_installed,
)


def run_fiducial_based_alignment(
        tilt_series_file: Path,
        tilt_angles: List[float],
        pixel_size: float,
        fiducial_size: float,
        nominal_rotation_angle: float,
        output_directory: Path,
):
    """Run fiducial based alignment in IMOD on a single tilt-series.

    Parameters
    ----------
    tilt_series_file: file containing tilt-series images.
        File must be compatible with the version of IMOD installed.
    tilt_angles: nominal stage tilt-angles from the microscope.
    pixel_size: nominal pixel size in Angstroms per pixel.
    fiducial_size: approximate size of fiducials in nanometers.
    nominal_rotation_angle: initial estimate for the rotation angle of the tilt
        axis. https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    output_directory: tilt-series directory for IMOD.
    """
    if not imod_is_installed():
        raise RuntimeError('No IMOD installation found.')

    prepare_imod_directory(
        tilt_series_file=tilt_series_file,
        tilt_angles=tilt_angles,
        imod_directory=output_directory
    )
    directive = generate_fiducial_alignment_directive(
        tilt_series_file=tilt_series_file,
        pixel_size=pixel_size,
        fiducial_size=fiducial_size,
        rotation_angle=nominal_rotation_angle
    )
    run_batchruntomo(
        tilt_series_file=tilt_series_file,
        imod_directory=output_directory,
        directive=directive
    )


def generate_fiducial_alignment_directive(
        tilt_series_file: PathLike,
        pixel_size: float,
        fiducial_size: float,
        rotation_angle: float
) -> Dict[str, Any]:
    """Generate a fiducial-based alignment directive file for batchruntomo

    Parameters
    ----------
    tilt_series_file : file containing the tilt-series stack
    pixel_size : pixel size in the tilt-series (angstroms per pixel)
    fiducial_size : fiducial size (nanometers)
    rotation_angle : initial estimate for the rotation angle
        https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    """
    alignment_binning_factor = find_optimal_power_of_2_binning_factor(
        src_pixel_size=pixel_size, target_pixel_size=TARGET_PIXEL_SIZE_FOR_ALIGNMENT
    )
    directive = read_adoc(BATCHRUNTOMO_CONFIG_FIDUCIALS)
    directive['setupset.copyarg.stackext'] = Path(tilt_series_file).suffix
    directive['setupset.copyarg.rotation'] = str(rotation_angle)
    directive['setupset.copyarg.pixel'] = str(pixel_size / 10)
    directive['setupset.copyarg.gold'] = str(fiducial_size)
    directive['comparam.prenewst.newstack.BinByFactor'] = str(alignment_binning_factor)
    return directive



import os
import subprocess
import tempfile
from pathlib import Path
from typing import List, Dict

import numpy as np

from yet_another_imod_wrapper.batchruntomo_config.io import write_adoc


def imod_is_installed() -> bool:
    command = ['batchruntomo', '--help']
    completed_process = subprocess.run(command)
    if completed_process.returncode == 0:
        return True
    return False


def prepare_imod_directory(
        tilt_series_file: Path, tilt_angles: List[float], imod_directory: Path
):
    root_name = tilt_series_file.stem
    imod_directory.mkdir(exist_ok=True, parents=True)

    tilt_series_file_for_imod = imod_directory / tilt_series_file.name
    os.symlink(tilt_series_file, tilt_series_file_for_imod)

    rawtlt_file = imod_directory / f'{root_name}.rawtlt'
    np.savetxt(rawtlt_file, tilt_angles, fmt='%.2f', delimiter='')


def run_batchruntomo(
        tilt_series_file: Path, imod_directory: Path, directive: Dict[str, str]
):
    root_name = tilt_series_file.stem
    with tempfile.TemporaryDirectory() as temporary_directory:
        directive_file = Path(temporary_directory) / 'directive.adoc'
        write_adoc(directive, directive_file)
        batchruntomo_command = [
            'batchruntomo',
            '-DirectiveFile', f'{directive_file}',
            '-RootName', f'{root_name}',
            '-CurrentLocation', f'{imod_directory}',
            '-EndingStep', '6'
        ]
        subprocess.run(batchruntomo_command)


def _find_optimal_binning_factor(
        binning_factors: np.ndarray,
        src_pixel_size: float,
        target_pixel_size: float
) -> int:
    binned_pixel_sizes = binning_factors * src_pixel_size
    pixel_size_deltas = np.abs(binned_pixel_sizes - target_pixel_size)
    return binning_factors[np.argmin(pixel_size_deltas)]


def find_optimal_integer_binning_factor(
        src_pixel_size: float, target_pixel_size: float
) -> int:
    binning_factors = np.arange(1, 30)
    return _find_optimal_binning_factor(binning_factors, src_pixel_size, target_pixel_size)


def find_optimal_power_of_2_binning_factor(
        src_pixel_size: float, target_pixel_size: float
) -> int:
    binning_factors = 2 ** np.arange(6)
    return _find_optimal_binning_factor(binning_factors, src_pixel_size, target_pixel_size)
